import os

import numpy as np
import SimpleITK as sitk

from asltk.utils import load_image, save_image


class ASLData:
    def __init__(
        self,
        **kwargs,
    ):
        """ASLData constructor

        The basic data need to represent a ASL data is the full path to load
        the image file, the Labeling Duration (LD) array and the Post-labeling
        Delay (PLD) array. Is none of those information is passed, a null
        ASLData object is created, which can be further been fed using the
        get/set methods.

        The constructor is generic for classic ASL data and also for multi-TE
        and Diffusion-Weighted (DW) ASL protocols. There is a specfic get/set
        method for TE/DW data. If TE/DW is not provided, then it is assumed as
        type `None` for those data properties. In order to informs the TE or DW
        values in the object instance, you can use the tags `te_values` or
        `dw_values` in the construction call

        Examples:
            By default, the LD and PLD arrays are indicated (as empty lists)

            >>> data = ASLData()
            >>> data.get_ld()
            []
            >>> data.get_pld()
            []

            >>> data = ASLData(te_values=[13.0,20.2,50.5,90.5,125.2])
            >>> data.get_te()
            [13.0, 20.2, 50.5, 90.5, 125.2]

            >>> data = ASLData(dw_values=[13.0,20.2,50.5,90.5,125.2])
            >>> data.get_dw()
            [13.0, 20.2, 50.5, 90.5, 125.2]

        Other parameters: Set the ASL data parameters
            pcasl (str, optional): The ASL data full path with filename. Defaults to ''.
            m0 (str, optional): The M0 data full path with filename. Defaults to ''.
            ld_values (list, optional): The LD values. Defaults to [].
            pld_values (list, optional): The PLD values. Defaults to [].
            te_values (list, optional): The TE values. Defaults to None.
            dw_values (list, optional): The DW values. Defaults to None.
        """
        self._asl_image = None
        self._m0_image = None
        self._parameters = {
            'ld': [],
            'pld': [],
            'te': None,
            'dw': None,
        }

        if kwargs.get('pcasl') is not None:
            self._asl_image = load_image(kwargs.get('pcasl'))

        if kwargs.get('m0') is not None:
            self._m0_image = load_image(kwargs.get('m0'))

        self._parameters['ld'] = (
            [] if kwargs.get('ld_values') is None else kwargs.get('ld_values')
        )
        self._parameters['pld'] = (
            []
            if kwargs.get('pld_values') is None
            else kwargs.get('pld_values')
        )
        self._check_ld_pld_sizes(
            self._parameters['ld'], self._parameters['pld']
        )
        if kwargs.get('te_values'):
            self._parameters['te'] = kwargs.get('te_values')
        if kwargs.get('dw_values'):
            self._parameters['dw'] = kwargs.get('dw_values')

    def set_image(self, full_path: str, spec: str):
        """Insert a image necessary to define de ASL data processing.

        The `spec` parameters specifies what is the type of image to be used in
        ASL processing step. Choose one of the options: `m0` for the M0 volume,
        `pcasl` for the pCASL data.

        Examples:
            >>> data = ASLData()
            >>> path_m0 = './tests/files/m0.nii.gz' # M0 file with shape (5,35,35)
            >>> data.set_image(path_m0, spec='m0')
            >>> data('m0').shape
            (5, 35, 35)

        Args:
            full_path (str): The full path with filename to be loaded
            spec (str): The type of image being used in the ASL processing.
        """
        match spec:
            case 'm0':
                self._m0_image = load_image(full_path)
            case 'pcasl':
                self._asl_image = load_image(full_path)

    def get_ld(self):
        """Obtain the LD array values"""
        return self._parameters['ld']

    def set_ld(self, ld_values: list):
        """Set the LD values.

        The proper way to inform the values here is using a list of int or
        float data. The total quantity of values depends on the image
        acquisition protocol.

        The list length for LD must be equal to PLD list length.

        Args:
            ld_values (list): The values to be adjusted for LD array
        """
        self._check_input_parameter(ld_values, 'LD')
        self._parameters['ld'] = ld_values

    def get_pld(self):
        """Obtain the PLD array values"""
        return self._parameters['pld']

    def set_pld(self, pld_values: list):
        """Set the PLD values.

        The proper way to inform the values here is using a list of int or
        float data. The total quantity of values depends on the image
        acquisition protocol.

        The list length for PLD must be equal to LD list length.

        Args:
            pld_values (list): The values to be adjusted for PLD array
        """
        self._check_input_parameter(pld_values, 'PLD')
        self._parameters['pld'] = pld_values

    def get_te(self):
        """Obtain the TE array values"""
        return self._parameters['te']

    def set_te(self, te_values: list):
        """Set the TE values.

        The proper way to inform the values here is using a list of int or
        float data. The total quantity of values depends on the image
        acquisition protocol.

        Args:
            te_values (list): The values to be adjusted for TE array
        """
        self._check_input_parameter(te_values, 'TE')
        self._parameters['te'] = te_values

    def get_dw(self):
        """Obtain the PLD array values"""
        return self._parameters['dw']

    def set_dw(self, dw_values: list):
        """Set the DW values.

        The proper way to inform the values here is using a list of int or
        float data. The total quantity of values depends on the image
        acquisition protocol.

        Args:
            dw_values (list): The values to be adjusted for DW array
        """
        self._check_input_parameter(dw_values, 'DW')
        self._parameters['dw'] = dw_values

    def __call__(self, spec: str):
        """Object caller to expose the image data.

        Examples:
            >>> data = ASLData(pcasl='./tests/files/t1-mri.nrrd')
            >>> type(data('pcasl'))
            <class 'numpy.ndarray'>

            >>> np.min(data('pcasl'))
            np.uint8(0)

        Returns:
            (numpy.ndarray): The data placed in the ASLData object
        """
        match spec:
            case 'pcasl':
                return self._asl_image
            case 'm0':
                return self._m0_image

    def _check_input_parameter(self, values, param_type):
        for v in values:
            if not isinstance(v, int) and not isinstance(v, float):
                raise ValueError(
                    f'{param_type} values is not a list of valid numbers.'
                )
            if v <= 0:
                raise ValueError(
                    f'{param_type} values must be postive non zero numbers.'
                )

    def _check_ld_pld_sizes(self, ld, pld):
        if len(ld) != len(pld):
            raise ValueError(
                f'LD and PLD must have the same array size. LD size is {len(ld)} and PLD size is {len(pld)}'
            )



import warnings

import numpy as np
from rich import print
from rich.progress import track
from scipy.optimize import curve_fit

from asltk.asldata import ASLData
from asltk.mri_parameters import MRIParameters
from asltk.utils import asl_model_buxton, asl_model_multi_te

# TODO Opcao para aplicar filtro no mapa de saida (Gauss, etc)
# TODO Brain mask como input ou opção de processamento?


class CBFMapping(MRIParameters):
    def __init__(self, asl_data: ASLData) -> None:
        """Basic CBFMapping constructor.

        Notes:
            The ASLData is the base data used in the object constructor.
            In order to create the CBF map correctly, a proper ASLData must be
            provided. Check whether the ASLData given as input is defined
            correctly

        Examples:
            The default MRIParameters are used as default in the object
            constructor
            >>> asl_data = ASLData(pcasl='./tests/files/pcasl.nii.gz',m0='./tests/files/m0.nii.gz')
            >>> cbf = CBFMapping(asl_data)
            >>> cbf.get_constant('T1csf')
            1400.0

            If the user want to change the MRIParameter value, for a specific
            object, one can change it directly:
            >>> cbf.set_constant(1600.0, 'T1csf')
            >>> cbf.get_constant('T1csf')
            1600.0
            >>> default_param = MRIParameters()
            >>> default_param.get_constant('T1csf')
            1400.0

        Args:
            asl_data (ASLData): The ASL data object (ASLData)
        """
        super().__init__()
        self._asl_data = asl_data
        if self._asl_data('m0') is None:
            raise ValueError(
                'ASLData is incomplete. CBFMapping need pcasl and m0 images.'
            )

        self._brain_mask = np.ones(self._asl_data('m0').shape)
        self._cbf_map = np.zeros(self._asl_data('m0').shape)
        self._att_map = np.zeros(self._asl_data('m0').shape)

    def set_brain_mask(self, brain_mask: np.ndarray, label: int = 1):
        """Defines whether a brain a mask is applied to the CBFMapping
        calculation

        A image mask is simply an image that defines the voxels where the ASL
        calculation should be made. Basically any integer value can be used as
        proper label mask.

        A most common approach is to use a binary image (zeros for background
        and 1 for the brain tissues). Anyway, the default behavior of the
        method can transform a integer-pixel values image to a binary mask with
        the `label` parameter provided by the user

        Args:
            brain_mask (np.ndarray): The image representing the brain mask label (int, optional): The label value used to define the foreground tissue (brain). Defaults to 1.
        """
        # TODO Add flag to create new mask using fsl.bet
        self._check_mask_values(brain_mask, label)

        binary_mask = (brain_mask == label).astype(np.uint8) * label
        self._brain_mask = binary_mask

    def create_map(
        self,
        ub: list = [1.0, 5000.0],
        lb: list = [0.0, 0.0],
        par0: list = [1e-5, 1000],
    ):
        """Create the CBF and also ATT maps

        Note:
            By default the ATT map is already calculated using the same Buxton
            formalism. Once the CBFMapping.create_map() method is called, both
            CBF and ATT maps are given in the output.

        Note:
            The CBF maps is given in two formats: the original pixel scale,
            resulted from the non-linear Buxton model fitting, and also
            a normalized version with the correct units of mL/100 g/min. In the
            output dictionary the user can select the 'cbf' and 'cbf_norm'
            options

        Args:
            ub (list, optional): The upper limit values. Defaults to [1.0, 5000.0].
            lb (list, optional): The lower limit values. Defaults to [0.0, 0.0].
            par0 (list, optional): The initial guess parameter for non-linear fitting. Defaults to [1e-5, 1000].

        Returns:
            (dict): A dictionary with 'cbf', 'att' and 'cbf_norm'
        """

        BuxtonX = [
            self._asl_data.get_ld(),
            self._asl_data.get_pld(),
        ]   # x data for the Buxton model

        x_axis = self._asl_data('m0').shape[2]   # height
        y_axis = self._asl_data('m0').shape[1]   # width
        z_axis = self._asl_data('m0').shape[0]   # depth

        for i in track(
            range(x_axis), description='[green]CBF/ATT processing...'
        ):
            for j in range(y_axis):
                for k in range(z_axis):
                    if self._brain_mask[k, j, i] != 0:
                        m0_px = self._asl_data('m0')[k, j, i]

                        def mod_buxton(Xdata, par1, par2):
                            return asl_model_buxton(
                                Xdata[0], Xdata[1], m0_px, par1, par2
                            )

                        # TODO precisa achar uma forma generica para saber estrutura asl_data (TE,PLD,3DVOL)... como o codigo irá puxar de forma direta?
                        Ydata = self._asl_data('pcasl')[
                            0, :, k, j, i
                        ]   # taking the first TE to get less influence in the PLD/Buxton model

                        # TODO Investigar mais...quando chama multiTE e ele chama o CBF acusa erro RuntimeError. Dado utilizado LOAMSample/pac2
                        try:
                            par_fit, _ = curve_fit(
                                mod_buxton,
                                BuxtonX,
                                Ydata,
                                p0=par0,
                                bounds=(lb, ub),
                            )
                            self._cbf_map[k, j, i] = par_fit[0]
                            self._att_map[k, j, i] = par_fit[1]
                        except RuntimeError:
                            self._cbf_map[k, j, i] = 0.0
                            self._att_map[k, j, i] = 0.0

        return {
            'cbf': self._cbf_map,
            'cbf_norm': self._cbf_map * (60 * 60 * 1000),
            'att': self._att_map,
        }

    def _check_mask_values(self, mask, label):
        # Check whether the mask provided is a binary image
        unique_values = np.unique(mask)
        if unique_values.size > 2:
            warnings.warn(
                'Mask image is not a binary image. Any value > 0 will be assumed as brain label.',
                UserWarning,
            )

        # Check whether the label value is found in the mask image
        label_ok = False
        for value in unique_values:
            if label == value:
                label_ok = True
                break
        if not label_ok:
            raise ValueError('Label value is not found in the mask provided.')

        # Check whether the dimensions between mask and input volume matches
        mask_shape = mask.shape
        input_vol_shape = self._asl_data('pcasl')[0, 0, :, :, :].shape
        if mask_shape != input_vol_shape:
            raise TypeError(
                f'Image mask dimension does not match with input 3D volume. Mask shape {mask_shape} not equal to {input_vol_shape}'
            )


class MultiTE_ASLMapping(MRIParameters):
    def __init__(self, asl_data: ASLData) -> None:
        super().__init__()
        self._asl_data = asl_data
        self._basic_maps = CBFMapping(asl_data)
        if self._asl_data('m0') is None:
            raise ValueError(
                'ASLData is incomplete. CBFMapping need pcasl and m0 images.'
            )

        self._brain_mask = np.ones(self._asl_data('m0').shape)
        self._cbf_map = np.zeros(self._asl_data('m0').shape)
        self._att_map = np.zeros(self._asl_data('m0').shape)
        self._t1blgm_map = np.zeros(self._asl_data('m0').shape)

    def set_brain_mask(self, brain_mask: np.ndarray, label: int = 1):
        """Defines whether a brain a mask is applied to the CBFMapping
        calculation

        A image mask is simply an image that defines the voxels where the ASL
        calculation should be made. Basically any integer value can be used as
        proper label mask.

        A most common approach is to use a binary image (zeros for background
        and 1 for the brain tissues). Anyway, the default behavior of the
        method can transform a integer-pixel values image to a binary mask with
        the `label` parameter provided by the user

        Args:
            brain_mask (np.ndarray): The image representing the brain mask label (int, optional): The label value used to define the foreground tissue (brain). Defaults to 1.
        """
        # TODO Add flag to create new mask using fsl.bet
        self._check_mask_values(brain_mask, label)

        binary_mask = (brain_mask == label).astype(np.uint8) * label
        self._brain_mask = binary_mask

    def set_cbf_map(self, cbf_map: np.ndarray):
        """Set the CBF map to the MultiTE_ASLMapping object.

        Note:
            The CBF maps must have the original scale in order to calculate the
            T1blGM map correclty. Hence, if the CBF map was made using
            CBFMapping class, one can use the 'cbf' output.

        Args:
            cbf_map (np.ndarray): The CBF map that is set in the MultiTE_ASLMapping object
        """
        self._cbf_map = cbf_map

    def get_cbf_map(self) -> np.ndarray:
        """Get the CBF map storaged at the MultiTE_ASLMapping object

        Returns:
            (np.ndarray): The CBF map that is storaged in the
            MultiTE_ASLMapping object
        """
        return self._cbf_map

    def set_att_map(self, att_map: np.ndarray):
        """Set the ATT map to the MultiTE_ASLMapping object.

        Args:
            att_map (np.ndarray): The ATT map that is set in the MultiTE_ASLMapping object
        """
        self._att_map = att_map

    def get_att_map(self):
        """Get the ATT map storaged at the MultiTE_ASLMapping object

        Returns:
            (np.ndarray): _description_
        """
        return self._att_map

    def _check_mask_values(self, mask, label):
        # Check whether the mask provided is a binary image
        unique_values = np.unique(mask)
        if unique_values.size > 2:
            warnings.warn(
                'Mask image is not a binary image. Any value > 0 will be assumed as brain label.',
                UserWarning,
            )

        # Check whether the label value is found in the mask image
        label_ok = False
        for value in unique_values:
            if label == value:
                label_ok = True
                break
        if not label_ok:
            raise ValueError('Label value is not found in the mask provided.')

        # Check whether the dimensions between mask and input volume matches
        mask_shape = mask.shape
        input_vol_shape = self._asl_data('pcasl')[0, 0, :, :, :].shape
        if mask_shape != input_vol_shape:
            raise TypeError(
                f'Image mask dimension does not match with input 3D volume. Mask shape {mask_shape} not equal to {input_vol_shape}'
            )

    def create_map(
        self,
        ub: list = [np.inf],
        lb: list = [0.0],
        par0: list = [400],
    ):
        """Create the T1 relaxation exchange between blood and Grey Matter (GM)
        , i.e. the T1blGM map resulted from the multi-compartiment TE ASL model
        Reference:  Ultra-long-TE arterial spin labeling reveals rapid and
        brain-wide blood-to-CSF water transport in humans, NeuroImage,
        doi: 10.1016/j.neuroimage.2021.118755

        Note:
            The CBF and ATT maps can be provided before calling this method,
            using the proper set/get methods for each map. If the user does not
            provide these maps, a new calculation is automatically made using
            the default execution implemented at CBFMapping class.

        Note:
            The CBF map must be at the original scale to perform the correct
            multiTE-ASL model. Therefore, provide the 'cbf' output.

        Args:
            ub (list, optional): The upper limit values. Defaults to [1.0, 5000.0].
            lb (list, optional): The lower limit values. Defaults to [0.0, 0.0].
            par0 (list, optional): The initial guess parameter for non-linear fitting. Defaults to [1e-5, 1000].

        Returns:
            (dict): A dictionary with 'cbf', 'att' and 'cbf_norm'
        """
        # TODO As entradas ub, lb e par0 não são aplicadas para CBF. Pensar se precisa ter essa flexibilidade para acertar o CBF interno à chamada
        cbf = CBFMapping(self._asl_data)
        cbf.set_brain_mask(self._brain_mask)

        basic_maps = {'cbf': self._cbf_map, 'att': self._att_map}
        # If the CBF/ATT maps are zero (empty), then a new one is created
        print(
            '[blue][INFO] The CBF/ATT map were not provided. Creating these maps before next step...'
        )
        if np.int8(np.mean(self._cbf_map)) == np.int8(0) or np.int8(
            np.mean(self._att_map)
        ) == np.int8(0):
            basic_maps = self._basic_maps.create_map()
            self._cbf_map = basic_maps['cbf']
            self._att_map = basic_maps['att']

        x_axis = self._asl_data('m0').shape[2]   # height
        y_axis = self._asl_data('m0').shape[1]   # width
        z_axis = self._asl_data('m0').shape[0]   # depth

        for i in track(
            range(x_axis), description='[green]multiTE-ASL processing...'
        ):
            for j in range(y_axis):
                for k in range(z_axis):
                    if self._brain_mask[k, j, i] != 0:
                        m0_px = self._asl_data('m0')[k, j, i]

                        def mod_2comp(Xdata, par1):
                            return asl_model_multi_te(
                                Xdata[:, 0],
                                Xdata[:, 1],
                                Xdata[:, 2],
                                m0_px,
                                basic_maps['cbf'][k, j, i],
                                basic_maps['att'][k, j, i],
                                par1,
                                self.T2bl,
                                self.T2gm,
                            )

                        # Ydata = np.swapaxes(asl_img[:,:,k,j,i],0,1).reshape((len(pld) * len(te), 1), order='F').flatten('F')
                        Ydata = (
                            self._asl_data('pcasl')[:, :, k, j, i]
                            .reshape(
                                (
                                    len(self._asl_data.get_ld())
                                    * len(self._asl_data.get_te()),
                                    1,
                                )
                            )
                            .flatten()
                        )

                        try:
                            Xdata = self._create_x_data(
                                self._asl_data.get_ld(),
                                self._asl_data.get_pld(),
                                self._asl_data.get_te(),
                            )
                            par_fit, _ = curve_fit(
                                mod_2comp,
                                Xdata,
                                Ydata,
                                p0=par0,
                                bounds=(lb, ub),
                            )
                            self._t1blgm_map[k, j, i] = par_fit[0]
                        except RuntimeError:
                            self._t1blgm_map[k, j, i] = 0.0

        return {
            'cbf': self._cbf_map,
            'cbf_norm': self._cbf_map * (60 * 60 * 1000),
            'att': self._att_map,
            't1blgm': self._t1blgm_map,
        }

    def _create_x_data(self, ld, pld, te):
        # array for the x values, assuming an arbitrary size based on the PLD
        # and TE vector size
        Xdata = np.zeros((len(pld) * len(te), 3))

        count = 0
        for i in range(len(pld)):
            for j in range(len(te)):
                Xdata[count] = [ld[i], pld[i], te[j]]
                count += 1

        return Xdata



import argparse
import os
from functools import *

import numpy as np
import SimpleITK as sitk
from rich import print
from rich.progress import track
from scipy.optimize import curve_fit

from asltk.asldata import ASLData
from asltk.reconstruction import CBFMapping
from asltk.utils import load_image, save_image

parser = argparse.ArgumentParser(
    description='Python script to calculate the basic CBF and ATT maps from ASL data.'
)

parser.add_argument(
    'pcasl',
    type=str,
    help='ASL raw data obtained from the MRI scanner. This must be the basic PLD ASL MRI acquisition protocol.',
)
parser.add_argument('m0', type=str, help='M0 image in Nifti format.')
parser.add_argument(
    'mask',
    type=str,
    nargs='?',
    default='',
    help='Image mask defining the ROI where the calculations must be done. Any pixel value different from zero will be assumed as the ROI area. Outside the mask (value=0) will be ignored.',
)
parser.add_argument(
    'out_folder',
    type=str,
    nargs='?',
    default=os.path.expanduser('~'),
    help='The output folder that is the reference to save all the output images in the script. The images selected to be saved are given as tags in the script caller, e.g. the options --cbf_map and --att_map. By default, the TblGM map is placed in the output folder with the name tblgm_map.nii.gz',
)
parser.add_argument(
    '--pld',
    type=str,
    nargs='+',
    required=True,
    help='Posts Labeling Delay (PLD) trend, arranged in a sequence of float numbers',
)
parser.add_argument(
    '--ld',
    type=str,
    nargs='+',
    required=True,
    help='Labeling Duration trend (LD), arranged in a sequence of float numbers.',
)
parser.add_argument(
    '--verbose',
    action='store_true',
    help='Show more details thoughout the processing.',
)

args = parser.parse_args()

# Script check-up parameters
def checkUpParameters():
    is_ok = True
    # Check output folder exist
    if not (os.path.isdir(args.out_folder)):
        print(
            f'Output folder path does not exist (path: {args.out_folder}). Please create the folder before executing the script.'
        )
        is_ok = False

    # Check ASL image exist
    if not (os.path.isfile(args.pcasl)):
        print(
            f'ASL input file does not exist (file path: {args.asl}). Please check the input file before executing the script.'
        )
        is_ok = False

    # Check M0  image exist
    if not (os.path.isfile(args.m0)):
        print(
            f'M0 input file does not exist (file path: {args.m0}). Please check the input file before executing the script.'
        )
        is_ok = False

    return is_ok


asl_img = load_image(args.pcasl)
m0_img = load_image(args.m0)

mask_img = np.ones(asl_img[0, 0, :, :, :].shape)
if args.mask != '':
    mask_img = load_image(args.mask)


try:
    pld = [float(s) for s in args.pld]
    ld = [float(s) for s in args.ld]
except:
    pld = [float(s) for s in str(args.pld[0]).split()]
    ld = [float(s) for s in str(args.ld[0]).split()]

if not checkUpParameters():
    raise RuntimeError(
        'One or more arguments are not well defined. Please, revise the script call.'
    )


# Step 2: Show the input information to assist manual conference
if args.verbose:
    print(' --- Script Input Data ---')
    print('ASL file path: ' + args.pcasl)
    print('ASL image dimension: ' + str(asl_img.shape))
    print('Mask file path: ' + args.mask)
    print('Mask image dimension: ' + str(mask_img.shape))
    print('M0 file path: ' + args.m0)
    print('M0 image dimension: ' + str(m0_img.shape))
    print('PLD: ' + str(pld))
    print('LD: ' + str(ld))

data = ASLData(pcasl=args.pcasl, m0=args.m0, ld_values=ld, pld_values=pld)
recon = CBFMapping(data)
recon.set_brain_mask(mask_img)
maps = recon.create_map()


save_path = args.out_folder + os.path.sep + 'cbf_map.nii.gz'
if args.verbose:
    print('Saving CBF map - Path: ' + save_path)
save_image(maps['cbf'], save_path)

save_path = args.out_folder + os.path.sep + 'cbf_map_normalized.nii.gz'
if args.verbose:
    print('Saving normalized CBF map - Path: ' + save_path)
save_image(maps['cbf_norm'], save_path)

save_path = args.out_folder + os.path.sep + 'att_map.nii.gz'
if args.verbose:
    print('Saving ATT map - Path: ' + save_path)
save_image(maps['att'], save_path)

if args.verbose:
    print('Execution: ' + parser.prog + ' finished successfully!')
